<?php

namespace Drupal\themespace\Plugin\Discovery;

use Doctrine\Common\Annotations\AnnotationRegistry;
use Drupal\Component\Annotation\AnnotationInterface;
use Drupal\Component\Annotation\Doctrine\SimpleAnnotationReader;
use Drupal\Component\Annotation\Doctrine\StaticReflectionParser;
use Drupal\Component\Annotation\Reflection\MockFileFinder;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Plugin\Discovery\AttributeClassDiscovery;
use Drupal\themespace\Annotation\ProviderTypedPlugin as AnnotationProviderTypedPlugin;
use Drupal\themespace\Annotation\ProviderTypedPluginInterface;
use Drupal\themespace\Attribute\ProviderTypedPlugin;

/**
 * Attribute class discovery for plugins which are aware of the provider type.
 *
 * A plugin discovery handler for plugin attributes which are aware of their
 * provider's extension type. For theme provided plugins, is it important
 * for plugin managers and consumers to be aware these are provided by themes
 * and should only be used when that theme is active.
 *
 * @see \Drupal\themespace\Attribute\ProviderTypedPluginInterface
 */
class ProviderTypedAttributeDiscoveryWithAnnotation extends ProviderTypedAttributeClassDiscovery implements ProviderTypedDiscoveryInterface {

  /**
   * The doctrine annotation reader.
   *
   * @var \Drupal\Component\Annotation\Doctrine\SimpleAnnotationReader|null
   */
  protected ?SimpleAnnotationReader $annotationReader;

  /**
   * Creates a new instance of the ProviderTypedAnnotatedClassDiscovery class.
   *
   * @param string $subdir
   *   Either the plugin's subdirectory, for example 'Plugin/views/filter', or
   *   empty string if plugins are located at the top level of the namespace.
   * @param \Traversable $root_namespaces
   *   An object that implements \Traversable which contains the root paths
   *   keyed by the corresponding namespace to look for plugin implementations.
   *   If $subdir is not an empty string, it will be appended to each namespace.
   * @param string $pluginDefinitionAttributeName
   *   (optional) The name of the annotation that contains the plugin
   *   definition. Defaults to 'Drupal\Component\Attribute\Plugin'.
   * @param string $pluginDefinitionAnnotationName
   *   (optional) The name of the attribute that contains the plugin definition.
   *   Defaults to 'Drupal\Component\Annotation\Plugin'.
   * @param string[] $additionalNamespaces
   *   (optional) Additional namespaces to scan for attribute definitions.
   */
  public function __construct(
    string $subdir,
    \Traversable $root_namespaces,
    string $pluginDefinitionAttributeName = ProviderTypedPlugin::class,
    protected readonly string $pluginDefinitionAnnotationName = AnnotationProviderTypedPlugin::class,
    protected readonly array $additionalNamespaces = [],
  ) {
    parent::__construct($subdir, $root_namespaces, $pluginDefinitionAttributeName);
  }

  /**
   * {@inheritdoc}
   */
  protected function getFileCacheSuffix(string $defaultSuffix):string {
    return $defaultSuffix . ':' . Crypt::hashBase64(serialize($this->additionalNamespaces)) . ':' . str_replace('\\', '_', $this->pluginDefinitionAnnotationName);
  }

  /**
   * Prepares the annotation definition.
   *
   * This is a copy of the prepareAnnotationDefinition method from annotated
   * class discovery.
   *
   * @param \Drupal\Component\Annotation\AnnotationInterface $annotation
   *   The annotation derived from the plugin.
   * @param class-string $class
   *   The class used for the plugin.
   *
   * @see \Drupal\Component\Annotation\Plugin\Discovery\AnnotatedClassDiscovery::prepareAnnotationDefinition()
   * @see \Drupal\Core\Plugin\Discovery\AnnotatedClassDiscovery::prepareAnnotationDefinition()
   */
  protected function prepareAnnotationDefinition(AnnotationInterface $annotation, $class): void {
    $annotation->setClass($class);

    if ($annotation instanceof ProviderTypedPluginInterface) {
      if (!$annotation->getProvider() || !$annotation->getProviderType()) {
        [$provider, $type] = $this->getProviderInfoFromNamespace($class);
        $annotation->setProvider($provider);
        $annotation->setProviderType($type);
      }
    }
    elseif (!$annotation->getProvider()) {
      $annotation->setProvider($this->getProviderFromNamespace($class));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getDefinitions(): array {
    // Clear the annotation loaders of any previous annotation classes.
    AnnotationRegistry::reset();
    // Register the namespaces of classes that can be used for annotations.
    // @phpstan-ignore-next-line
    AnnotationRegistry::registerLoader('class_exists');

    // Ensure this is reset before attempting to do any discovery.
    $this->rootNamespacesIterator = $this->fullNamespaceIterator;

    $definitions = AttributeClassDiscovery::getDefinitions();
    $this->annotationReader = NULL;

    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getModuleDefinitions(): array {
    // Clear the annotation loaders of any previous annotation classes.
    AnnotationRegistry::reset();
    // Register the namespaces of classes that can be used for annotations.
    // @phpstan-ignore-next-line
    AnnotationRegistry::registerLoader('class_exists');

    // Create an iterator that only traverses namespaces that match the Drupal
    // namespace pattern. We alter the $this->rootNamespacesIterator so it can
    // be use in the parent::getDefinitions().
    // This avoids having to reimplement
    // self::getDefinitions().
    $iterator = new \IteratorIterator($this->fullNamespaceIterator);
    $this->rootNamespacesIterator = new \CallbackFilterIterator($iterator, function ($value, $key) {
      return !str_starts_with($key, 'Drupal\\Theme\\');
    });

    $definitions = AttributeClassDiscovery::getDefinitions();
    $this->annotationReader = NULL;

    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getThemeDefinitions(): array {
    // Clear the annotation loaders of any previous annotation classes.
    AnnotationRegistry::reset();
    // Register the namespaces of classes that can be used for annotations.
    // @phpstan-ignore-next-line
    AnnotationRegistry::registerLoader('class_exists');

    // @see self::getModuleDefinitions()
    $iterator = new \IteratorIterator($this->fullNamespaceIterator);
    $this->rootNamespacesIterator = new \CallbackFilterIterator($iterator, function ($value, $key) {
      return str_starts_with($key, 'Drupal\\Theme\\');
    });

    $definitions = AttributeClassDiscovery::getDefinitions();
    $this->annotationReader = NULL;

    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  protected function parseClass(string $class, \SplFileInfo $fileinfo): array {
    // The filename is already known, so there is no need to find the
    // file. However, StaticReflectionParser needs a finder, so use a
    // mock version.
    $finder = MockFileFinder::create($fileinfo->getPathName());
    $parser = new StaticReflectionParser($class, $finder, TRUE);

    /** @var \Drupal\Component\Annotation\Doctrine\StaticReflectionClass $reflectClass */
    $reflectClass = $parser->getReflectionClass();
    /** @var \Drupal\Component\Annotation\AnnotationInterface $annotation */
    $annotation = $this->getAnnotationReader()->getClassAnnotation($reflectClass, $this->pluginDefinitionAnnotationName);
    if ($annotation) {
      $this->prepareAnnotationDefinition($annotation, $class);
      return ['id' => $annotation->getId(), 'content' => $annotation->get()];
    }

    // Annotations use static reflection and are able to analyze a class that
    // extends classes or uses traits that do not exist. Attribute discovery
    // will trigger a fatal error with such classes, so only call it if the
    // class has a class attribute.
    if ($reflectClass->hasClassAttribute($this->pluginDefinitionAttributeName)) {
      return parent::parseClass($class, $fileinfo);
    }
    return ['id' => NULL, 'content' => NULL];
  }

  /**
   * Gets the used doctrine annotation reader.
   *
   * This is a copy of the getAnnotationReader method from annotated class
   * discovery.
   *
   * @return \Drupal\Component\Annotation\Doctrine\SimpleAnnotationReader
   *   The annotation reader.
   *
   * @see \Drupal\Component\Annotation\Plugin\Discovery\AnnotatedClassDiscovery::getAnnotationReader()
   * @see \Drupal\Core\Plugin\Discovery\AnnotatedClassDiscovery::getAnnotationReader()
   */
  private function getAnnotationReader() : SimpleAnnotationReader {
    if (!isset($this->annotationReader)) {
      $this->annotationReader = new SimpleAnnotationReader();

      // Add the namespaces from the main plugin annotation, like @EntityType.
      $namespace = substr($this->pluginDefinitionAnnotationName, 0, strrpos($this->pluginDefinitionAnnotationName, '\\'));
      $this->annotationReader->addNamespace($namespace);

      // Register additional namespaces to be scanned for annotations.
      foreach ($this->additionalNamespaces as $namespace) {
        $this->annotationReader->addNamespace($namespace);
      }

      // Add the Core annotation classes like @Translation.
      $this->annotationReader->addNamespace('Drupal\Core\Annotation');
    }
    return $this->annotationReader;
  }

}
